# -*- coding: utf-8 -*-
"""module for coordinate manipulation (conversions, calculations etc.)

(c) 2007-2009 Matt Hilton 

U{http://astlib.sourceforge.net}

"""

import sys
import math
from PyWCSTools import wcscon

#---------------------------------------------------------------------------------------------------
def hms2decimal(RAString, delimiter):
    """Converts a delimited string of Hours:Minutes:Seconds format into decimal degrees. 
    
    @type RAString: string
    @param RAString: coordinate string in H:M:S format
    @type delimiter: string
    @param delimiter: delimiter character in RAString
    @rtype: float
    @return: coordinate in decimal degrees
    
    """
    # is it in HH:MM:SS format?
    if delimiter=="":
        RABits=str(RAString).split()
    else:
        RABits=str(RAString).split(delimiter)
    if len(RABits)>1:
        RAHDecimal=float(RABits[0])
        if len(RABits)>1:
            RAHDecimal=RAHDecimal+(float(RABits[1])/60.0)
        if len(RABits)>2:
            RAHDecimal=RAHDecimal+(float(RABits[2])/3600.0)
        RADeg=(RAHDecimal/24.0)*360.0
    else:
        RADeg=float(RAString)

    return RADeg

#---------------------------------------------------------------------------------------------------
def dms2decimal(decString, delimiter):
    """Converts a delimited string of Degrees:Minutes:Seconds format into decimal degrees.
    
    @type decString: string
    @param decString: coordinate string in D:M:S format
    @type delimiter: string
    @param delimiter: delimiter character in decString
    @rtype: float
    @return: coordinate in decimal degrees
        
    """
    # is it in DD:MM:SS format?
    if delimiter=="":
        decBits=str(decString).split()
    else:
        decBits=str(decString).split(delimiter)
    if len(decBits)>1:
        decDeg=float(decBits[0])
        if decBits[0].find("-")!=-1:
            if len(decBits)>1:
                decDeg=decDeg-(float(decBits[1])/60.0)
            if len(decBits)>2:
                decDeg=decDeg-(float(decBits[2])/3600.0)
        else:
            if len(decBits)>1:
                decDeg=decDeg+(float(decBits[1])/60.0)
            if len(decBits)>2:
                decDeg=decDeg+(float(decBits[2])/3600.0)
    else:
        decDeg=float(decString)
    
    return decDeg

#---------------------------------------------------------------------------------------------------
def decimal2hms(RADeg, delimiter):
    """Converts decimal degrees to string in Hours:Minutes:Seconds format with user specified
    delimiter.
    
    @type RADeg: float
    @param RADeg: coordinate in decimal degrees
    @type delimiter: string
    @param delimiter: delimiter character in returned string
    @rtype: string
    @return: coordinate string in H:M:S format

    
    """
    hours=(RADeg/360.0)*24
    if hours<10 and hours>=1:
        sHours="0"+str(hours)[0]
    elif hours>=10:
        sHours=str(hours)[:2]
    elif hours<1:
        sHours="00"
        
    # Python turns small numbers into strings like e.g. 5e-5, so if no '.' found, must be small
    if str(hours).find(".")==-1:
        mins=float(hours)*60.0
    else:
        mins=float(str(hours)[str(hours).index("."):])*60.0
    if mins<10 and mins>=1:
        sMins="0"+str(mins)[:1]
    elif mins>=10:
        sMins=str(mins)[:2]
    elif mins<1:
        sMins="00"
        
    secs=(hours-(float(sHours)+float(sMins)/60.0))*3600.0
    if secs<10 and secs>0.001:
        sSecs="0"+str(secs)[:str(secs).find(".")+4]
# 	elif secs<1:
# 		sSecs="00"+str(secs)[:str(secs).find(".")+3]
    elif secs<0.0001:
        sSecs="00.001"
    else:
        sSecs=str(secs)[:str(secs).find(".")+4] 
    if len(sSecs)<5:
        sSecs=sSecs+"00"	# So all to 3dp
        
    
    return sHours+delimiter+sMins+delimiter+sSecs

#---------------------------------------------------------------------------------------------------
def decimal2dms(decDeg, delimiter):
    """Converts decimal degrees to string in Degrees:Minutes:Seconds format with user specified
    delimiter.
    
    @type decDeg: float
    @param decDeg: coordinate in decimal degrees
    @type delimiter: string
    @param delimiter: delimiter character in returned string
    @rtype: string
    @return: coordinate string in D:M:S format
    
    """
    # Positive
    if decDeg>0:
        if decDeg<10 and decDeg>=1:
            sDeg="0"+str(decDeg)[0]
        elif decDeg>=10:
            sDeg=str(decDeg)[:2]
        elif decDeg<1:
            sDeg="00"
        
        if str(decDeg).find(".")==-1:
            mins=float(decDeg)*60.0
        else:
            mins=float(str(decDeg)[str(decDeg).index("."):])*60
        if mins<10 and mins>=1:
            sMins="0"+str(mins)[:1]
        elif mins>=10:
            sMins=str(mins)[:2]
        elif mins<1:
            sMins="00"
            
        secs=(decDeg-(float(sDeg)+float(sMins)/60.0))*3600.0
        if secs<10 and secs>0:
            sSecs="0"+str(secs)[:str(secs).find(".")+3]
        elif secs<0.001:
            sSecs="00.00"
        else:
            sSecs=str(secs)[:str(secs).find(".")+3]
        if len(sSecs)<5:
            sSecs=sSecs+"0"	# So all to 2dp
    
        return "+"+sDeg+delimiter+sMins+delimiter+sSecs
    
    else:
        if decDeg>-10 and decDeg<=-1:
            sDeg="-0"+str(decDeg)[1]
        elif decDeg<=-10:
            sDeg=str(decDeg)[:3]
        elif decDeg>-1:
            sDeg="-00"
            
        if str(decDeg).find(".")==-1:
            mins=float(decDeg)*-60.0
        else:
            mins=float(str(decDeg)[str(decDeg).index("."):])*60			
        if mins<10 and mins>=1:
            sMins="0"+str(mins)[:1]
        elif mins>=10:
            sMins=str(mins)[:2]
        elif mins<1:
            sMins="00"
            
        secs=(decDeg-(float(sDeg)-float(sMins)/60.0))*3600.0
        if secs>-10 and secs<0:
            sSecs="0"+str(secs)[1:str(secs).find(".")+3]	# so don't get minus sign
        elif secs>-0.001:
            sSecs="00.00"
        else:
            sSecs=str(secs)[1:str(secs).find(".")+3]
        if len(sSecs)<5:
            sSecs=sSecs+"0"	# So all to 2dp	
                    
        return sDeg+delimiter+sMins+delimiter+sSecs

#---------------------------------------------------------------------------------------------------
def calcAngSepDeg(RADeg1, decDeg1, RADeg2, decDeg2):
    """Calculates the angular separation of two positions on the sky (specified in decimal
    degrees) in decimal degrees, assuming a tangent plane projection (so separation has to be
    <90 deg).
    
    @type RADeg1: float
    @param RADeg1: R.A. in decimal degrees for position 1
    @type decDeg1: float
    @param decDeg1: dec. in decimal degrees for position 1
    @type RADeg2: float
    @param RADeg2: R.A. in decimal degrees for position 2
    @type decDeg2: float
    @param decDeg2: dec. in decimal degrees for position 2
    @rtype: float
    @return: angular separation in decimal degrees	
    
    """
    cRA=math.radians(float(RADeg1))
    cDec=math.radians(float(decDeg1))
    
    gRA=math.radians(float(RADeg2))
    gDec=math.radians(float(decDeg2))
            
    dRA=cRA-gRA
    dDec=gDec-cDec
    cosC=(math.sin(gDec)*math.sin(cDec))+(math.cos(gDec)*math.cos(cDec)*math.cos(gRA-cRA))
    x=(math.cos(cDec)*math.sin(gRA-cRA))/cosC
    y=((math.cos(gDec)*math.sin(cDec))-(math.sin(gDec)*math.cos(cDec)*math.cos(gRA-cRA)))/cosC
    r=math.degrees(math.sqrt(x*x+y*y))
    
    return r
    
#---------------------------------------------------------------------------------------------------
def convertCoords(inputSystem, outputSystem, coordX, coordY, epoch):	
    """Converts specified coordinates (given in decimal degrees) between J2000, B1950, and
    Galactic.
    
    @type inputSystem: string
    @param inputSystem: system of the input coordinates (either "J2000", "B1950" or "GALACTIC")
    @type outputSystem: string
    @param outputSystem: system of the returned coordinates (either "J2000", "B1950" or
    "GALACTIC")
    @type coordX: float
    @param coordX: longitude coordinate in decimal degrees, e.g. R. A.
    @type coordY: float
    @param coordY: latitude coordinate in decimal degrees, e.g. dec.
    @type epoch: float
    @param epoch: epoch of the input coordinates
    @rtype: list
    @return: coordinates in decimal degrees in requested output system
    
    
    """
    
    if inputSystem=="J2000" or inputSystem=="B1950" or inputSystem=="GALACTIC":
        if outputSystem=="J2000" or outputSystem=="B1950" or outputSystem=="GALACTIC":
            
            outCoords=wcscon.wcscon(wcscon.wcscsys(inputSystem),
                wcscon.wcscsys(outputSystem), 0, 0, coordX, coordY, epoch)
            
            return outCoords

    raise Exception, "inputSystem and outputSystem must be 'J2000', 'B1950' or 'GALACTIC'"
        
#---------------------------------------------------------------------------------------------------
    