# -*- coding: utf-8 -*-
"""module for producing astronomical plots

(c) 2007-2009 Matt Hilton 

U{http://astlib.sourceforge.net}

This module provides the matplotlib powered ImagePlot class, which is designed to be flexible. 
ImagePlots can have RA, Dec. coordinate axes, contour overlays, and have objects marked in them, 
using WCS coordinates. RGB plots are supported too.

@var DEC_TICK_STEPS: Defines the possible coordinate label steps on the delination axis in
sexagesimal mode. Dictionary format: {'deg', 'int', 'unit'}
@type DEC_TICK_STEPS: dictionary list

@var RA_TICK_STEPS: Defines the possible coordinate label steps on the right ascension axis in
sexagesimal mode. Dictionary format: {'deg', 'int', 'unit'}
@type RA_TICK_STEPS: dictionary list

@var DECIMAL_TICK_STEPS: Defines the possible coordinate label steps on both coordinate axes in
decimal degrees mode.
@type DECIMAL_TICK_STEPS: list

@var DEG: Variable to stand in for the degrees symbol.
@type DEG: string

@var PRIME: Variable to stand in for the prime symbol.
@type PRIME: string

@var DOUBLE_PRIME: Variable to stand in for the double prime symbol.
@type DOUBLE_PRIME: string

"""

import math
import astImages
import astWCS
import astCoords
import numpy
import pyfits
try:
    import pylab
    import matplotlib.patches as patches
except:
    print "WARNING: astPlots: failed to import matplotlib - some functions will not work."
import sys

DEC_TICK_STEPS=[{'deg': 1.0/60.0/60.0,  'int': 1,   'unit': "s"}, 
                {'deg': 5.0/60.0/60.0,  'int': 5,   'unit': "s"}, 
                {'deg': 10.0/60.0/60.0, 'int': 10,  'unit': "s"},
                {'deg': 30.0/60.0/60.0, 'int': 30,  'unit': "s"},
                {'deg': 1.0/60.0,       'int': 1,   'unit': "m"},
                {'deg': 5.0/60.0,       'int': 5,   'unit': "m"},
                {'deg': 15.0/60.0,      'int': 15,  'unit': "m"},
                {'deg': 30.0/60.0,      'int': 30,  'unit': "m"}, 
                {'deg': 1.0,            'int': 1,    'unit': "d"},
                {'deg': 5.0,            'int': 5,    'unit': "d"},
                {'deg': 10.0,           'int': 10,   'unit': "d"},
                {'deg': 30.0,           'int': 30,   'unit': "d"}]

RA_TICK_STEPS=[ {'deg': (1.0/60.0/60.0/24.0)*360.0,  'int': 1,   'unit': "s"}, 
                {'deg': (5.0/60.0/60.0/24.0)*360.0,  'int': 5,   'unit': "s"}, 
                {'deg': (10.0/60.0/60.0/24.0)*360.0, 'int': 10,  'unit': "s"},
                {'deg': (30.0/60.0/60.0/24.0)*360.0, 'int': 30,  'unit': "s"},
                {'deg': (1.0/60.0/24.0)*360.0,       'int': 1,   'unit': "m"},
                {'deg': (5.0/60.0/24.0)*360.0,       'int': 5,   'unit': "m"},
                {'deg': (10.0/60.0/24.0)*360.0,      'int': 15,  'unit': "m"},
                {'deg': (30.0/60.0/24.0)*360.0,      'int': 30,  'unit': "m"}, 
                {'deg': (1.0/24.0)*360.0,            'int': 1,   'unit': "h"},
                {'deg': (3.0/24.0)*360.0,            'int': 3,   'unit': "h"},
                {'deg': (6.0/24.0)*360.0,            'int': 6,   'unit': "h"},
                {'deg': (12.0/24.0)*360.0,           'int': 12,  'unit': "h"}]

DECIMAL_TICK_STEPS=[0.001, 0.0025, 0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.5, 5.0, 10.0, 30.0, 90.0]

DEG = u"\N{DEGREE SIGN}"
PRIME = "\'"            #u"\N{PRIME}" ought to work, but not with my fonts
DOUBLE_PRIME = "\""     #u"\N{DOUBLE PRIME}"

#---------------------------------------------------------------------------------------------------
class ImagePlot:
    """This class describes a matplotlib image plot containing an astronomical image with an
    associated WCS.
    
    Objects within the image boundaries can be marked by passing their WCS coordinates to 
    L{ImagePlot.addPlotObjects}.
    
    Other images can be overlaid using L{ImagePlot.addContourOverlay}.
    
    For images rotated with North at the top, East at the left (as can be done using
    L{astImages.clipRotatedImageSectionWCS}), WCS coordinate axes can be plotted, with tic marks
    set appropriately for the image size (note that this may not currently be 100% reliable for
    sexagesimal coordinates). Otherwise, a compass can be plotted showing the directions of North
    and East in the image.

    RGB images are also supported.
    
    The plot can of course be tweaked further after creation using matplotlib/pylab commands.
    
    """
    def __init__(self, imageData, imageWCS, axes = [0.1,0.1,0.8,0.8], \
        cutLevels = ["smart", 99.5], colorMapName = "gray", title = None, axesLabels = "decimal", \
        minorLabels = False, axesFontFamily="serif", axesFontSize=12.0, colorBar = False):
        """Makes an ImagePlot from the given image array and astWCS. For coordinate axes to work, the
        image and WCS should have been rotated such that East is at the left, North is at the top
        (see e.g. L{astImages.clipRotatedImageSectionWCS}).
        
        If imageData is given as a list in the format [r, g, b], an color RGB plot will be made. However,
        in this case the cutLevels must be specified manually for each component as a list -
        i.e. cutLevels = [[r min, r max], [g min, g max], [b min, b max]]. In this case of course, the
        colorMap will be ignored. All r, g, b image arrays must have the same dimensions.
        
        Set axesLabels = None to make a plot without coordinate axes plotted.
        
        The axes can be marked in either sexagesimal or decimal celestial coordinates: the appropriate 
        axis scales will be determined automatically from the size of the image array and associated
        WCS.
        
        @type imageData: numpy array or list
        @param imageData: image data array or list of numpy arrays [r, g, b]
        @type imageWCS: astWCS.WCS
        @param imageWCS: astWCS.WCS object
        @type axes: list
        @param axes: specifies where in the current figure to draw the finder chart (see pylab.axes)
        @type cutLevels: list
        @param cutLevels: sets the image scaling - available options:
            - pixel values: cutLevels=[low value, high value].
            - histogram equalisation: cutLevels=["histEq", number of bins ( e.g. 1024)]
            - relative: cutLevels=["relative", cut per cent level (e.g. 99.5)]
            - smart: cutLevels=["smart", cut per cent level (e.g. 99.5)]
        ["smart", 99.5] seems to provide good scaling over a range of different images.
        Note that for RGB images, cut levels must be specified manually i.e. as a list:
        [[r min, rmax], [g min, g max], [b min, b max]]
        @type colorMapName: string
        @param colorMapName: name of a standard matplotlib colormap, e.g. "hot", "cool", "gray"
        etc. (do "help(pylab.colormaps)" in the Python interpreter to see available options)
        @type title: string
        @param title: optional title for the plot
        @type axesLabels: string
        @param axesLabels: either "sexagesimal" (for H:M:S, D:M:S), "decimal" (for decimal degrees)
        or None (for no coordinate axes labels)
        @type minorLabels: bool
        @param minorLabels: if set to True, add additional labels between coordinate labels
        @type axesFontFamily: string
        @param axesFontFamily: matplotlib fontfamily, e.g. 'serif', 'sans-serif' etc.
        @type axesFontSize: float
        @param axesFontSize: font size of axes labels and titles (in points)
        @type colorBar: bool
        @param colorBar: if True, plot a vertical color bar at the side of the image indicating the intensity
        scale.
        
        """
                
        self.RADeg, self.decDeg=imageWCS.getCentreWCSCoords()
        self.wcs=imageWCS
        
        # Handle case where imageData is [r, g, b]
        if type(imageData) == list:
            if len(imageData) == 3:
                if len(cutLevels) == 3:
                    r=astImages.normalise(imageData[0], cutLevels[0])
                    g=astImages.normalise(imageData[1], cutLevels[1])
                    b=astImages.normalise(imageData[2], cutLevels[2])
                    rgb=numpy.array([r.transpose(), g.transpose(), b.transpose()])
                    rgb=rgb.transpose()
                    self.data=rgb
                    self.rgbImage=True
                else:
                    raise Exception, "tried to create a RGB array, but cutLevels is not a list of 3 lists"

            else:
                raise Exception, "tried to create a RGB array but imageData is not a list of 3 arrays"
        else:
            self.data=imageData
            self.rgbImage=False
        
        self.axes=pylab.axes(axes)
        self.cutLevels=cutLevels
        self.colorMapName=colorMapName
        self.title=title
        self.axesLabels=axesLabels
        self.minorLabels=minorLabels
        self.colorBar=colorBar
        self.axesFontSize=axesFontSize
        self.axesFontFamily=axesFontFamily
        
        if self.axesLabels != None:
            self.calcWCSAxisLabels(axesLabels = self.axesLabels, minorLabels = self.minorLabels)
        
        # this list stores objects to overplot, add to it using addPlotObjects()
        self.plotObjects=[] 
        
        # this list stores image data to overlay as contours, add to it using addContourOverlay()
        self.contourOverlays=[]
        
        self.draw()


    def draw(self):
        """Redraws the ImagePlot - this should be called after adding plot objects, compass, contours
        etc.
        
        """
        
        pylab.axes(self.axes)
        pylab.cla()
        
        if self.title != None:
            pylab.title(self.title)
        try:
            colorMap=pylab.cm.get_cmap(self.colorMapName)
        except AssertionError:
            raise Exception, self.colorMapName+"is not a defined matplotlib colormap."
        
        if self.rgbImage == False:
            self.cutImage=astImages.intensityCutImage(self.data, self.cutLevels)
            if self.cutLevels[0]=="histEq":
                pylab.imshow(self.cutImage['image'],  interpolation="bilinear", origin='lower', cmap=colorMap)
            else:
                pylab.imshow(self.cutImage['image'],  interpolation="bilinear",  norm=self.cutImage['norm'], \
                            origin='lower', cmap=colorMap)
        else:
            pylab.imshow(self.data, interpolation="bilinear", origin='lower')
        
        if self.colorBar == True:
            pylab.colorbar(shrink=0.8)
        
        for c in self.contourOverlays:
            pylab.contour(c['contourData']['scaledImage'], c['contourData']['contourLevels'], 
                            colors=c['color'], linewidths=c['width'])
        
        for p in self.plotObjects:
            for x, y, l in zip(p['x'], p['y'], p['objLabels']):
                if p['symbol'] == "circle":
                    c=patches.Circle((x, y), radius=p['sizePix']/2.0, fill=False, edgecolor=p['color'], 
                                        linewidth=p['width'])
                    self.axes.add_patch(c)
                elif p['symbol'] == "box":
                    c=patches.Rectangle((x-p['sizePix']/2, y-p['sizePix']/2), p['sizePix'], p['sizePix'], 
                        fill=False, edgecolor=p['color'], linewidth=p['width'])
                    self.axes.add_patch(c)
                elif p['symbol'] == "cross":
                    pylab.plot([x-p['sizePix']/2, x+p['sizePix']/2], [y, y], linestyle='-', 
                        linewidth=p['width'], color= p['color'])
                    pylab.plot([x, x], [y-p['sizePix']/2, y+p['sizePix']/2], linestyle='-', 
                        linewidth=p['width'], color= p['color'])
                if l != None:
                    pylab.text(x, y+p['sizePix']/1.5, l, horizontalalignment='center', \
                                fontsize=p['objLabelSize'], color=p['color'])
            
            if p['symbol'] == "compass":
                x=p['x'][0]
                y=p['y'][0]
                ra=p['RA'][0]
                dec=p['dec'][0]
                northPoint=dec+p['sizeArcSec']/3600.0
                eastPoint=ra+p['sizeArcSec']/3600.0
                sizePix=(p['sizeArcSec']/3600.0)/self.wcs.getPixelSizeDeg()
                northPix=self.wcs.wcs2pix(ra, northPoint)
                eastPix=self.wcs.wcs2pix(eastPoint, dec)
                edx=eastPix[0]-x
                edy=eastPix[1]-y
                ndx=northPix[0]-x
                ndy=northPix[1]-y
                nArrow=patches.Arrow(x, y, ndx, ndy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])
                eArrow=patches.Arrow(x, y, edx, edy, edgecolor=p['color'], facecolor=p['color'], width=p['width'])                
                self.axes.add_patch(nArrow)
                self.axes.add_patch(eArrow)
                pylab.text(x+ndx+ndx*0.2, y+ndy+ndy*0.2, "N", horizontalalignment='center', 
                                verticalalignment='center', fontsize=p['objLabelSize'], color=p['color'])
                pylab.text(x+edx+edx*0.2, y+edy+edy*0.2, "E", horizontalalignment='center', 
                                verticalalignment='center', fontsize=p['objLabelSize'], color=p['color'])

        if self.axesLabels != None:
            pylab.xticks(self.ticsRA[0], self.ticsRA[1], weight='normal', family=self.axesFontFamily, \
                                    fontsize=self.axesFontSize)
            pylab.yticks(self.ticsDec[0], self.ticsDec[1], weight='normal', family=self.axesFontFamily, \
                                    fontsize=self.axesFontSize)
            pylab.xlabel(self.RAAxisLabel, family=self.axesFontFamily, fontsize=self.axesFontSize)
            pylab.ylabel(self.decAxisLabel, family=self.axesFontFamily, fontsize=self.axesFontSize)
        else:
            pylab.xticks([], [])
            pylab.yticks([], [])
            pylab.xlabel("")
            pylab.ylabel("")
                
        pylab.xlim(0, self.data.shape[1]-1)
        pylab.ylim(0, self.data.shape[0]-1)


    def addContourOverlay(self, contourImageData, contourWCS, label, levels = ['linear', 'min', 'max', 5], \
                             width = 1, color = "white", smooth = 0):
        """Adds image data to the ImagePlot as a contour overlay. The contours will be plotted when
        ImagePlot.draw() is next called. The contours can be removed using L{removeContourOverlay}
        If a contour overlay already exists with this label, it will be replaced.
        
        @type contourImageData: numpy array
        @param contourImageData: image data array from which contours are to be generated
        @type contourWCS: astWCS.WCS
        @param contourWCS: astWCS.WCS object for the image to be contoured
        @type label: string
        @param label: identifying label for this set of contours
        @type levels: list
        @param levels: sets the contour levels - available options:
            - values: contourLevels=[list of values specifying each level]
            - linear spacing: contourLevels=['linear', min level value, max level value, number
            of levels]
            - log spacing: contourLevels=['log', min level value, max level value, number of
            levels]
        @type width: int
        @param width: width of the overlaid contours
        @type color: string
        @param color: color of the overlaid contours, specified by the name of a standard
            matplotlib color, e.g., "black", "white", "cyan"
            etc. (do "help(pylab.colors)" in the Python interpreter to see available options)
        @type smooth: float
        @param smooth: standard deviation (in pixels) of Gaussian filter for
            pre-smoothing of contour image data (set to 0 for no smoothing)
        
        """
                
        if self.rgbImage == True:
            backgroundData=self.data[:,:,0]
        else:
            backgroundData=self.data
        contourData=astImages.generateContourOverlay(backgroundData, self.wcs, contourImageData, \
                                                        contourWCS, levels, smooth)
        
        alreadyGot=False
        for c in self.contourOverlays:
            if c['label'] == label:
                c['contourData']=contourData
                c['label']=label
                c['color']=color
                c['width']=width
                alreadyGot=True
                
        if alreadyGot == False:
            self.contourOverlays.append({'contourData': contourData, 'label': label, 'color': color, \
                                            'width': width})

    
    def removeContourOverlay(self, label):
        """Removes the contourOverlay from the ImagePlot corresponding to the label. The plot must be redrawn
        for the change to take effect.
        
        @type label: string
        @param label: label for contour overlay in ImagePlot.contourOverlays to be removed
        
        """
        
        index=0
        for p in self.contourOverlays:
            if p['label'] == label:
                self.plotObjects.remove(self.plotObjects[index])
            index=index+1
        
        
    def addPlotObjects(self, objRAs, objDecs, label, symbol="circle", size=4.0, width=1.0, color="yellow",                                           
                                    objLabels = None, objLabelSize = 12.0):
        """Add objects with RA, dec coords objRAs, objDecs to the ImagePlot. The objects will be plotted
        when ImagePlot.draw() is next called; only objects that fall within the image boundaries will be 
        plotted. 
        
        symbol specifies the type of symbol with which to mark the object in the image. The following
        values are allowed:
            - "circle"
            - "box"
            - "cross"
        
        size specifies the diameter in arcsec of the symbol (if plotSymbol == "circle"), or the width
        of the box in arcsec (if plotSymbol == "box")
        
        width specifies the thickness of the symbol lines in pixels
        
        color can be any valid matplotlib color (e.g. "red", "green", etc.)
        
        The objects can be removed from the plot by using removePlotObjects(), and then calling
        draw(). If the ImagePlot already has a set of plotObjects with the same label, they will be 
        replaced.
        
        @type objRAs: numpy array or list
        @param objRAs: object RA coords in decimal degrees
        @type objDecs: numpy array or list
        @param objDecs: corresponding object Dec. coords in decimal degrees
        @type label: string
        @param label: identifying label for this set of objects
        @type symbol: string
        @param symbol: either "circle" or "box"
        @type size: float
        @param size: size of symbols to plot (radius in arcsec, or width of box)
        @type width: float
        @param width: width of symbols in pixels
        @type color: string
        @param color: any valid matplotlib color string, e.g. "red", "green" etc.
        @type objLabels: list
        @param objLabels: text labels to plot next to objects in figure
        @type objLabelSize: float
        @param objLabelSize: size of font used for object labels (in points)
        
        """
        
        pixCoords=self.wcs.wcs2pix(objRAs, objDecs)
        
        xMax=self.data.shape[1]
        yMax=self.data.shape[0]
        
        if objLabels == None:
            objLabels=[None]*len(objRAs)
            
        xInPlot=[]
        yInPlot=[]
        RAInPlot=[]
        decInPlot=[]
        labelInPlot=[]
        for p, r, d, l in zip(pixCoords, objRAs, objDecs, objLabels):
            if p[0] >= 0 and p[0] < xMax and p[1] >= 0 and p[1] < yMax:
                xInPlot.append(p[0])
                yInPlot.append(p[1])
                RAInPlot.append(r)
                decInPlot.append(d)
                labelInPlot.append(l)
        
        xInPlot=numpy.array(xInPlot)
        yInPlot=numpy.array(yInPlot)
        RAInPlot=numpy.array(RAInPlot)
        decInPlot=numpy.array(decInPlot)
        
        # Size of symbols in pixels in plot - converted from arcsec
        sizePix=(size/3600.0)/self.wcs.getPixelSizeDeg()
        
        alreadyGot=False
        for p in self.plotObjects:
            if p['label'] == label:
                p['x']=xInPlot
                p['y']=yInPlot
                p['RA']=RAInPlot
                p['dec']=decInPlot
                p['label']=label
                p['objLabels']=objLabels
                p['symbol']=symbol
                p['sizePix']=sizePix
                p['sizeArcSec']=size
                p['width']=width
                p['color']=color
                p['objLabelSize']=objLabelSize
                alreadyGot=True
        
        if alreadyGot == False:
            self.plotObjects.append({'x': xInPlot, 'y': yInPlot, 'RA': RAInPlot, 'dec': decInPlot,
                                'label': label, 'objLabels': labelInPlot, 'symbol': symbol, 
                                'sizePix': sizePix, 'width': width, 'color': color,
                                'objLabelSize': objLabelSize, 'sizeArcSec': size})
        
        
    def removePlotObjects(self, label):
        """Removes the plotObjects from the ImagePlot corresponding to the label. The plot must be redrawn
        for the change to take effect.
        
        @type label: string
        @param label: label for set of objects in ImagePlot.plotObjects to be removed
        
        """
        
        index=0
        for p in self.plotObjects:
            if p['label'] == label:
                self.plotObjects.remove(self.plotObjects[index])
            index=index+1
    
        
    def addCompass(self, location, sizeArcSec, color = "white", fontSize = 12, \
                        width = 20.0):
        """Adds a compass to the ImagePlot at the given location ('N', 'NE', 'E', 'SE', 'S', 
        'SW', 'W', or 'NW'). Note these aren't directions on the WCS coordinate grid, they are 
        relative positions on the plot - so N is top centre, NE is top right, SW is bottom right etc.. 
        Alternatively, pixel coordinates (x, y) in the image can be given.
        
        @type location: string or tuple
        @param location: location in the plot where the compass is drawn:
            - string: N, NE, E, SE, S, SW, W or NW
            - tuple: (x, y)
        @type sizeArcSec: float
        @param sizeArcSec: length of the compass arrows on the plot in arc seconds
        @type color: string
        @param color: any valid matplotlib color string
        @type fontSize: float
        @param fontSize: size of font used to label N and E, in points
        @type width: float
        @param width: width of arrows used to mark compass
        
        """
        
        # Work out where the compass is going in WCS coords from the relative location given
        # Draw the compass 1.5 * the given size in arcsec from that edge of the plot
        if type(location) == str:
            RADeg, decDeg=self.wcs.getCentreWCSCoords()
            x, y=self.wcs.wcs2pix(RADeg, decDeg)
            halfWidthRADeg, halfHeightDecDeg=self.wcs.getHalfSizeDeg()
            compassDistRADeg=halfWidthRADeg-1.5*sizeArcSec/3600.0
            compassDistDecDeg=halfHeightDecDeg-1.5*sizeArcSec/3600.0
            if self.wcs.isFlipped() == True:
                compassDistRADeg=compassDistRADeg*-1
            foundLocation=False
            if location.find("N") != -1:
                y=y+compassDistDecDeg/self.wcs.getPixelSizeDeg()
                foundLocation=True
            if location.find("S") != -1:
                y=y-compassDistDecDeg/self.wcs.getPixelSizeDeg()
                foundLocation=True
            if location.find("E") != -1:
                x=x-compassDistRADeg/self.wcs.getPixelSizeDeg()
                foundLocation=True
            if location.find("W") != -1:
                x=x+compassDistRADeg/self.wcs.getPixelSizeDeg()
                foundLocation=True
            if foundLocation == False:
                raise Exception, "didn't understand location string for compass (should be e.g. N, S, E, W)."
        elif type(location) == tuple:
            x, y=location
        else:
            raise Exception, "didn't understand location for compass - should be string or tuple."
        RADeg, decDeg=self.wcs.pix2wcs(x, y)
        
        # Size of compass arrows pixels in plot - converted from arcsec
        sizePix=(sizeArcSec/3600.0)/self.wcs.getPixelSizeDeg()
        
        alreadyGot=False
        for p in self.plotObjects:
            if p['label'] == "compass":
                p['x']=[x]
                p['y']=[y]
                p['RA']=[RADeg]
                p['dec']=[decDeg]
                p['label']="compass"
                p['objLabels']=[None]
                p['symbol']="compass"
                p['sizePix']=sizePix
                p['sizeArcSec']=sizeArcSec
                p['width']=width
                p['color']=color
                p['objLabelSize']=fontSize
                alreadyGot=True
        
        if alreadyGot == False:
            self.plotObjects.append({'x': [x], 'y': [y], 'RA': [RADeg], 'dec': [decDeg],
                                'label': "compass", 'objLabels': [None], 'symbol': "compass", 
                                'sizePix': sizePix, 'width': width, 'color': color,
                                'objLabelSize': fontSize, 'sizeArcSec': sizeArcSec})
             
        
    def calcWCSAxisLabels(self, axesLabels = "decimal", minorLabels = False):
        """
        This function calculates the positions of coordinate labels for the RA and Dec axes of the 
        ImagePlot. The ImagePlot must be redrawn for changes to be applied.
        
        @type axesLabels: string
        @param axesLabels: either "sexagesimal" (for H:M:S, D:M:S), "decimal" (for decimal degrees),
        or None for no coordinate axes labels
        @type minorLabels: bool
        @param minorLabels: if set to True, add additional labels between coordinate labels
        
        """
        
        self.axesLabels=axesLabels
        self.minorLabels=minorLabels
        
        # Label equinox on axes
        equinox=self.wcs.getEquinox()
        if equinox<1984:
            equinoxLabel="B"+str(int(equinox))
        else:
            equinoxLabel="J"+str(int(equinox))
        
        # Work out axes labels
        topLeft=self.wcs.pix2wcs(0,0)
        bottomRight=self.wcs.pix2wcs(self.data.shape[1], self.data.shape[0])
        
        # Pick appropriate scale for tics 
        ticLabelListRA=[]
        ticLocListRA=[]
        ticLabelListDec=[]
        ticLocListDec=[]
            
        # Count up in RA sec from bottom right
        hmsRALeft=astCoords.decimal2hms(topLeft[0], ":")
        hmsRARight=astCoords.decimal2hms(bottomRight[0], ":")
        if topLeft[0] > bottomRight[0]:
            raDegDiff=abs(topLeft[0]-bottomRight[0])
        else:
            # handle wraparound
            raDegDiff=(360.0-bottomRight[0])+topLeft[0]
        tickMarkSteps=getTickSteps(raDegDiff, "ra")
        bitsRight=hmsRARight.split(":")
        bitsLeft=hmsRALeft.split(":")
        raMajorTickStep=tickMarkSteps['major']
        raMinorTickStep=tickMarkSteps['minor']
        raUnmarkedTickStep=tickMarkSteps['unlabelled']
        
        # Count up in dec arcsec from bottom right
        dmsDecTop=astCoords.decimal2dms(topLeft[1], ":")
        dmsDecBottom=astCoords.decimal2dms(bottomRight[1], ":")
        decDegDiff=abs(topLeft[1]-bottomRight[1])
        tickMarkSteps=getTickSteps(decDegDiff, "dec")
        bitsTop=dmsDecTop.split(":")
        bitsBottom=dmsDecBottom.split(":")
        decMajorTickStep=tickMarkSteps['major']
        decMinorTickStep=tickMarkSteps['minor']
        decUnmarkedTickStep=tickMarkSteps['unlabelled']
            
        if self.axesLabels == "sexagesimal":
            
            xAxisLabel="R.A. ("+equinoxLabel+")"
            yAxisLabel="Dec. ("+equinoxLabel+")"
            
            hRange=range(int(bitsRight[0])-2, int(bitsLeft[0])+2)
            # handle wrap around
            if bitsLeft[0] < bitsRight[0]:
                hRange=range(int(bitsRight[0]), 24)+range(int(bitsLeft[0]),int(bitsLeft[0])+1) 
            mRange=range(int(bitsRight[1]), int(bitsLeft[1])+2)
            if len(hRange)>1:
                mRange=range(0, 60)
            sRange=range(0, 60)
            for h in hRange:
                hString=str(h)
                if abs(h)<10:
                    hString="0"+hString
                for m in mRange:
                    mString=str(m)
                    if m<10:
                        mString="0"+str(m)
                    for s in sRange:
                        sString=str(s)+".000"
                        if s<10:
                            sString="0"+str(s)+".000"
                        hmsMark=hString+":"+mString+":"+sString
                        degMark=astCoords.hms2decimal(hmsMark, ":")
                        pixMark=self.wcs.wcs2pix(degMark, bottomRight[1])
                        if pixMark[0] > 0 and pixMark[0] < self.data.shape[1]:
                            addedLabel=False
                            if eval(raMajorTickStep['unit']+" % "+str(raMajorTickStep['int'])) == 0:
                                if raMajorTickStep['unit'] == "h" and m == 0 and s == 0 :
                                    label=hmsMark.split(":")[0]+".0"
                                    ticLabelListRA.append(label)
                                    ticLocListRA.append(int(round(pixMark[0])))
                                    addedLabel=True
                                elif raMajorTickStep['unit'] == "m" and s == 0:
                                    label=hmsMark.split(":")[0]+":"+hmsMark.split(":")[1]+".0"
                                    ticLabelListRA.append(label)
                                    ticLocListRA.append(int(round(pixMark[0])))
                                    addedLabel=True
                                elif raMajorTickStep['unit'] == "s":
                                    label=hmsMark.split(":")[0]+":"+hmsMark.split(":")[1]+":"+hmsMark.split(":")[2][:-4]+".0"
                                    ticLabelListRA.append(label)
                                    ticLocListRA.append(int(round(pixMark[0])))
                                    addedLabel=True
                                    
                            if self.minorLabels == True:
                                if addedLabel == False and eval(raMinorTickStep['unit']+" % "+str(raMinorTickStep['int'])) == 0:
                                    if raMinorTickStep['unit'] == "h" and m == 0 and s == 0 :
                                        label=""    # No minor tick marks if we're in hours
                                        ticLabelListRA.append(label)
                                        ticLocListRA.append(int(round(pixMark[0])))
                                        addedLabel=True
                                    elif raMinorTickStep['unit'] == "m" and s == 0:
                                        label=hmsMark.split(":")[1]+".0"
                                        ticLabelListRA.append(label)
                                        ticLocListRA.append(int(round(pixMark[0])))
                                        addedLabel=True
                                    elif raMinorTickStep['unit'] == "s":
                                        label=hmsMark.split(":")[2][:-4]+".0"
                                        ticLabelListRA.append(label)
                                        ticLocListRA.append(int(round(pixMark[0])))
                                        addedLabel=True
                                
                            if addedLabel == False and eval(raUnmarkedTickStep['unit']+" % "+str(raUnmarkedTickStep['int'])) == 0:
                                ticLabelListRA.append("")
                                ticLocListRA.append(int(round(pixMark[0])))
            
            # Dec.
            dRange=range(int(bitsBottom[0])-2, int(bitsTop[0])+2)
            if 0 in dRange:
                dRange.insert(dRange.index(0), 0)
                minusZero=True
            mRange=range(int(bitsBottom[1]), int(bitsTop[1])+2)
            if len(dRange)>1:
                mRange=range(0, 60)
            sRange=range(0, 60)
            for d in dRange:
                if d<0:
                    dString=str(d)
                elif d==0:
                    if minusZero == True:
                        dString="-"+str(d)
                        minusZero=False
                    else:
                        dString="+"+str(d)
                else:
                    dString="+"+str(d)
                if abs(d)<10:
                    dString=dString[:1]+"0"+dString[1:]
    
                for m in mRange:
                    mString=str(m)
                    if m<10:
                        mString="0"+str(m)
                    for s in sRange:
                        sString=str(s)+".00"
                        if s<10:
                            sString="0"+str(s)+".00"
                        dmsMark=dString+":"+mString+":"+sString
                        degMark=astCoords.dms2decimal(dmsMark, ":")
                        pixMark=self.wcs.wcs2pix(bottomRight[0], degMark)
                        if pixMark[1] > 0 and pixMark[1] < self.data.shape[0]:
                            addedLabel=False
                            if eval(decMajorTickStep['unit']+" % "+str(decMajorTickStep['int'])) == 0:
                                if decMajorTickStep['unit'] == "d" and m == 0 and s == 0 :
                                    if dmsMark.split(":")[0] == "-00":
                                        continue
                                    else:
                                        label=dmsMark.split(":")[0]+DEG+".0"
                                        ticLabelListDec.append(label)
                                        ticLocListDec.append(int(round(pixMark[1])))
                                        addedLabel=True
                                elif decMajorTickStep['unit'] == "m" and s == 0:
                                    if dmsMark.split(":")[0] == "-00" and dmsMark.split(":")[1] == "00":
                                        continue
                                    else:
                                        label=dmsMark.split(":")[0]+DEG+dmsMark.split(":")[1]+PRIME+".0"
                                        ticLabelListDec.append(label)
                                        ticLocListDec.append(int(round(pixMark[1])))
                                        addedLabel=True
                                elif decMajorTickStep['unit'] == "s":
                                    label=dmsMark.split(":")[0]+DEG+dmsMark.split(":")[1]+PRIME+dmsMark.split(":")[2][:-3]+DOUBLE_PRIME+".0"
                                    ticLabelListDec.append(label)
                                    ticLocListDec.append(int(round(pixMark[1])))
                                    addedLabel=True
    
                            if self.minorLabels == True:
                                if addedLabel == False and eval(decMinorTickStep['unit']+" % "+str(decMinorTickStep['int'])) == 0:
                                    if decMinorTickStep['unit'] == "d" and m == 0 and s == 0 :
                                        label=""    # No minor tick marks if we're in degrees
                                        ticLabelListDec.append(label)
                                        ticLocListDec.append(int(round(pixMark[1])))
                                        addedLabel=True
                                    elif decMinorTickStep['unit'] == "m" and s == 0:
                                        label=dmsMark.split(":")[1]+PRIME+".0"
                                        ticLabelListDec.append(label)
                                        ticLocListDec.append(int(round(pixMark[1])))
                                        addedLabel=True
                                    elif decMinorTickStep['unit'] == "s":
                                        label=dmsMark.split(":")[2][:-3]+DOUBLE_PRIME+".0"
                                        ticLabelListDec.append(label)
                                        ticLocListDec.append(int(round(pixMark[1])))
                                        addedLabel=True
                                
                            if addedLabel == False and eval(decUnmarkedTickStep['unit']+" % "+str(decUnmarkedTickStep['int'])) == 0:
                                ticLabelListDec.append("")
                                ticLocListDec.append(int(round(pixMark[1])))
        
        elif self.axesLabels == "decimal":
            
            xAxisLabel="R.A. Degrees ("+equinoxLabel+")"
            yAxisLabel="Dec. Degrees ("+equinoxLabel+")"
            
            if topLeft[0] > bottomRight[0]:
                raDegDiff=abs(topLeft[0]-bottomRight[0])
            else:
                # handle wraparound
                raDegDiff=(360.0-bottomRight[0])+topLeft[0]
            tickMarkSteps=getTickSteps(raDegDiff, None, mode = "decimal")
            raMajorTickStep=tickMarkSteps['major']
            raMinorTickStep=tickMarkSteps['minor']
            raUnmarkedTickStep=tickMarkSteps['unlabelled']
            raMajorRatio=raMajorTickStep/raUnmarkedTickStep
            raMinorRatio=raMinorTickStep/raUnmarkedTickStep
            raMajorFormat="%."+str(len(str(raMajorTickStep)[str(raMajorTickStep).find(".")+1:]))+"f"
            raMinorFormat="%."+str(len(str(raMinorTickStep)[str(raMinorTickStep).find(".")+1:]))+"f"
            
            # Note we're using same tick marks steps for both RA and dec (looks a bit neater)
            decDegDiff=abs(topLeft[1]-bottomRight[1])
            decMajorTickStep=tickMarkSteps['major']
            decMinorTickStep=tickMarkSteps['minor']
            decUnmarkedTickStep=tickMarkSteps['unlabelled']
            decMajorRatio=decMajorTickStep/decUnmarkedTickStep
            decMinorRatio=decMinorTickStep/decUnmarkedTickStep
            decMajorFormat="%."+str(len(str(decMajorTickStep)[str(decMajorTickStep).find(".")+1:]))+"f"
            decMinorFormat="%."+str(len(str(decMinorTickStep)[str(decMinorTickStep).find(".")+1:]))+"f"
            
            # don't want to have different labels at different precision
            if self.minorLabels == True:
                raMajorFormat=raMinorFormat
                decMajorFormat=decMinorFormat
                
            # Need to check this out at RA wrap around (probably be okay but slow)
            raMin=int(topLeft[0])-2
            raMax=int(bottomRight[0])+2
            raSteps=int((raMax-raMin)/raUnmarkedTickStep)
            
            decMin=int(bottomRight[1])-2
            decMax=int(topLeft[1])+2
            decSteps=int((decMax-decMin)/decUnmarkedTickStep)
    
            for i in range(raSteps):
                degMark=raMin+i*raUnmarkedTickStep
                if degMark < 360:
                    pixMark=self.wcs.wcs2pix(degMark, bottomRight[1])
                    if pixMark[0] > 0 and pixMark[0] < self.data.shape[1]:
                        addedLabel=False
                        if i % raMajorRatio == 0:
                            ticLabelListRA.append(raMajorFormat % degMark)
                            ticLocListRA.append(int(round(pixMark[0])))
                            addedLabel=True
                        
                        if self.minorLabels == True:
                            if addedLabel == False and i % raMinorRatio == 0:
                                ticLabelListRA.append(raMinorFormat % degMark)
                                ticLocListRA.append(int(round(pixMark[0])))
                                addedLabel=True
                        
                        if addedLabel == False:
                            ticLabelListRA.append("")
                            ticLocListRA.append(int(round(pixMark[0])))
    
            for i in range(decSteps):
                degMark=decMin+i*decUnmarkedTickStep
                pixMark=self.wcs.wcs2pix(bottomRight[0], degMark)
                if pixMark[1] > 0 and pixMark[1] < self.data.shape[0]:
                    addedLabel=False
                    if i % decMajorRatio == 0:
                        ticLabelListDec.append(decMajorFormat % degMark)
                        ticLocListDec.append(int(round(pixMark[1])))
                        addedLabel=True
                    
                    if self.minorLabels == True:
                        if addedLabel == False and i % decMinorRatio == 0:
                            ticLabelListDec.append(decMinorFormat % degMark)
                            ticLocListDec.append(int(round(pixMark[1])))
                            addedLabel=True
                    
                    if addedLabel == False:
                        ticLabelListDec.append("")
                        ticLocListDec.append(int(round(pixMark[1])))
    
        self.ticsRA=[ticLocListRA, ticLabelListRA]
        self.ticsDec=[ticLocListDec, ticLabelListDec]
        self.RAAxisLabel=xAxisLabel
        self.decAxisLabel=yAxisLabel


    def save(self, fileName):
        """Saves the ImagePlot in any format that matplotlib can understand, as determined from the 
        fileName extension.
        
        @type fileName: string
        @param fileName: path where plot will be written
        
        """
        
        pylab.draw()
        pylab.savefig(fileName)
        
        
#---------------------------------------------------------------------------------------------------
def getTickSteps(plotSizeDeg, axis, mode = "sexagesimal"):
    """Chooses the appropriate WCS coordinate tick steps for the given axis of a plot of given size.
    
    @type plotSizeDeg: float
    @param plotSizeDeg: size of plot in decimal degrees
    @type axis: string
    @param axis: either "ra" or "dec"
    @type mode: string
    @param mode: either "sexagesimal" (for H:M:S, D:M:S) or "decimal" (for decimal degrees)
    @rtype: dictionary
    @return: tick step sizes for major, minor and unlabelled plot ticks, in format {'major', 'minor',
    'unmarked'}
    
    @note: axis is ignored is mode = "decimal"
    
    """
    
    # Aim for 5 major tick marks on a plot
    
    if mode == "sexagesimal":
        
        if axis == "ra":
            matchIndex = 0
            for i in range(len(RA_TICK_STEPS)):
                if plotSizeDeg/3.0 > RA_TICK_STEPS[i]['deg']:
                    matchIndex = i
            # So don't wrap around the unmarked ticks
            if matchIndex<2:
                matchIndex=2
            
            return {'major': RA_TICK_STEPS[matchIndex], 'minor': RA_TICK_STEPS[matchIndex-1], 'unlabelled': RA_TICK_STEPS[matchIndex-2]}
            
        elif axis == "dec":
            matchIndex = 0
            for i in range(len(DEC_TICK_STEPS)):
                if plotSizeDeg/3.0 > DEC_TICK_STEPS[i]['deg']:
                    matchIndex = i
            # So don't wrap around the unmarked ticks
            if matchIndex<2:
                matchIndex=2
            
            return {'major': DEC_TICK_STEPS[matchIndex], 'minor': DEC_TICK_STEPS[matchIndex-1], 'unlabelled': DEC_TICK_STEPS[matchIndex-2]}
        
        else:
            raise Exception, "axis must be 'ra' or 'dec'"
    
    elif mode == "decimal":
        
            matchIndex = 0
            for i in range(len(DECIMAL_TICK_STEPS)):
                if plotSizeDeg/3.0 > DECIMAL_TICK_STEPS[i]:
                    matchIndex = i
            # So don't wrap around the unmarked ticks
            if matchIndex<2:
                matchIndex=2
            
            return {'major': DECIMAL_TICK_STEPS[matchIndex], 'minor': DECIMAL_TICK_STEPS[matchIndex-1], 'unlabelled': DECIMAL_TICK_STEPS[matchIndex-2]}
    
    else:
        raise Exception, "mode must be either 'sexagesimal' or 'decimal'"

#---------------------------------------------------------------------------------------------------
